/* rmask2repeats.c -- simplify RepeatMasker output
 * 
 * The "-genbank" option writes the output in GenBank header format.
 * The "-nosimple" option skips Simple/Low/Satellite repeat types.
 * The "-check" option checks endpoints against the given sequence length.
 * 
 * This program was requested by Mike Wilson.
 *
 * $Revision: 1.8 $$Date: 2001/04/03 17:03:02 $
 */

#include "util.h"

#define SYNTAX  "repeatmasker_file [-genbank] [-nosimple] " \
                  "[-check seq_length] > output_file"
#define DELIM   " \t\n"

enum { NAME, TYPE, BUFSIZE = 1000 };

struct foo {
	int field;
	const char *substr;
	const char *val;
} Rpt[] = {
	/* If the string in col #2 is a substring of the field named by col #1,
	 * then its PipMaker name is in col #3.  Use the first case that holds.
	 */
	{ NAME, "Alu",       "Alu" },
	{ NAME, "MIR",       "MIR" },
	{ TYPE, "L2",        "LINE2" },
	{ TYPE, "L1",        "LINE1" },
	{ TYPE, "LTR",       "LTR" },
	{ NAME, "LTR",       "LTR" },
	{ NAME, "HERV",      "LTR" },
	{ TYPE, "DNA",       "DNA" },
	{ NAME, "B1",        "B1" },
	{ TYPE, "B2",        "B2" },
	{ TYPE, "SINE",      "SINE" },
	{ TYPE, "LINE",      "Other" },
	{ NAME, "MML",       "Other" },
	{ NAME, "BUR1",      "Other" },
	{ TYPE, "Other",     "Other" },
	{ TYPE, "Unknown",   "Other" },
	{ TYPE, "RNA",       "RNA" },
	{ TYPE, "Simple",    "Simple" },
	{ TYPE, "Low",       "Simple" },
	{ TYPE, "Satellite", "Simple" }
};

static void get_args (int argc, char **argv, char **filename, bool *genbank,
                      bool *simple, int *seq_len);

int main (int argc, char **argv) {
	char *filename = NULL;                  /* name of RepeatMasker file    */
	bool genbank = 0;                       /* use GenBank header format?   */
	bool simple = 1;                        /* include simple repeats?      */
	int  seq_len = 0;                       /* max position for check       */
	FILE *fp;                               /* ptr to input file            */
	char buf [BUFSIZE];                     /* input buffer                 */
	char line [BUFSIZE];                    /* working copy of input line   */
	char *p;                                /* ptr to token                 */
	int start, end;                         /* endpoints of repeat          */
	char dir, *name, *type;                 /* other fields of repeat       */
	int nl = 0, nr = 0;                     /* count lines, repeats         */
	int tsize = sizeof (Rpt) / sizeof (Rpt[0]);   /* size of table          */
	int i;                                  /* table index                  */

	set_argv0 (argv[0]);
	get_args (argc, argv, &filename, &genbank, &simple, &seq_len);
	if (!filename)
		fatal (usage (SYNTAX));
	fp = ckopen (filename, "r");
	if (!genbank)
		printf ("%%:repeats\n");

/*
  0   1   2   3     4 5  6       7 8   9       10    11 12 13
413 5.6 0.0 0.0 HUMAN 1 54 (92195) C Alu SINE/Alu (238) 62 9
*/
	while (fgets (buf, BUFSIZE, fp)) {
		nl++;

		/* expect field[0] to be an integer */
		for (p = buf; *p == ' '; p++)
			;
		if (!isdigit (*p)) continue;

		/* skip field[0..4] */
		strcpy (line, buf);
		if (!(p = strtok (line, DELIM))) continue;
		if (!(p = strtok (0, DELIM))) continue;
		if (!(p = strtok (0, DELIM))) continue;
		if (!(p = strtok (0, DELIM))) continue;
		if (!(p = strtok (0, DELIM))) continue;

		/* expect field[5] and field[6] to be integers */
		if (!(p = strtok (0, DELIM))) continue;
		if (sscanf (p, "%d", &start) != 1)
			fatalf ("Could not convert start-point:\n%s", buf);

		if (!(p = strtok (0, DELIM))) continue;
		if (sscanf (p, "%d", &end) != 1)
			fatalf ("Could not convert end-point:\n%s", buf);

		if (start < 1 || start >= end)
			fatalf ("Address too low or out of order:\n%s", buf);
		if (seq_len > 0 && end > seq_len)
			fatalf ("Address exceeds seq_length of %d:\n%s", seq_len, buf);

		/* skip field[7] */
		if (!(p = strtok (0, DELIM))) continue;

		/* expect field[8] to be "+" or "C" */
		if (!(p = strtok (0, DELIM))) continue;
		dir = *p;
		if (dir != '+' && dir != 'C')
			fatalf ("Improper format of RepeatMasker file:\n%s", buf);

		/* expect field[9] to be a name */
		if (!(p = strtok (0, DELIM))) continue;
		name = p;

		/* expect field[10] to be a type */
		if (!(p = strtok (0, DELIM))) continue;
		type = p;

		nr++;

		/* print genbank-style output */
		if (genbank) {
			printf ("     repeat_region   ");
			if (dir == '+')
				printf ("%d..%d\n", start, end);
			else
				printf ("complement(%d..%d)\n", start, end);
			printf ("                     /rpt_family=\"%s\"\n", name);
			continue;
		}

		/* ... or regular output */
		for (i = 0; i < tsize; i++) {
			if ((Rpt[i].field == NAME && strstr (name, Rpt[i].substr)) ||
			    (Rpt[i].field == TYPE && strstr (type, Rpt[i].substr)))
				break;
		}
		if (i == tsize) {
			fprintf (stderr, "Unknown repeat at %d-%d, ", start, end);
			fprintf (stderr, "name = %s and type = %s\n", name, type);
		}
		else if (strsame (Rpt[i].val, "Simple")) {
			if (simple)
				printf ("%d %d %s\n", start, end, Rpt[i].val);
			else
				fprintf (stderr, "Skipping simple repeat at %d-%d\n",
				         start, end);
		}
		else {
			printf ("%d %d %s %s\n", start, end,
			        (dir == '+' ? "Right" : "Left"), Rpt[i].val);
		}
	}

	fclose (fp);
	if (nl && !nr)
		fatal ("No repeat elements were found.");
	return 0;
}

/* retrieve command-line arguments */
static void get_args (int argc, char **argv, char **filename, bool *genbank,
                      bool *simple, int *seq_len) {
	int i;

	for (i = 1; i < argc; i++) {
		if (strsame (argv[i], "-genbank"))
			*genbank = 1;
		else if (strsame (argv[i], "-nosimple"))
			*simple = 0;
		else if (strsame (argv[i], "-check")) {
			if (!(*seq_len = atoi (optval (argc, argv, &i, SYNTAX))))
				fatal ("Invalid length value.");
		}
		else if (!*filename)
			*filename = argv[i];
		else
			fatal (usage (SYNTAX));
	}
}
